import { world, system, Vector } from "@minecraft/server";
import "./config/loader";

let config = {
	"rns:stamina": "true",
	"rns:stamina_exhausted": "1",
	"rns:stamina_cooldown": "1",
	"rns:roll_cooldown": "1",
	"rns:stamina_strength": "312",
	"rns:roll": "true",
	"rns:roll_vertical_strength": "10",
	"rns:roll_horizontal_strength": "10",
	"rns:fatigue": "true",
	"rns:slow_damage": "true",
	"rns:roll_style": "1",
	"rns:stamina_style": "1",
	"rns:speedup": "true",
	"rns:edge_climb": "true"
};

function getActor(){
	let actor;
	for(let actor_type of world.getDimension("minecraft:overworld").getEntitiesAtBlockLocation({x:0, y:320, z:0})){
		if(actor_type.typeId == "kc:config_loader"){
			actor = actor_type;
			return actor;
		}
	}
	return actor;
}

function getConfigFromActor(){
	let l = {};
	let actor = getActor();
	if(actor == undefined) return {};
	for(let a of actor.getTags()){
		let b = a.split("=");
		l[b[0]] = b[1];
	}
	return l;
}

let refresh_ui = true;

system.afterEvents.scriptEventReceive.subscribe( s => {
	config = getConfigFromActor();
	refresh_ui = true;
}, { namespaces: [ "config_loaded" ]});

// world.afterEvents.worldInitialize.subscribe( s => {
// 	let data = new DynamicPropertiesDefinition();
// 	data.defineNumber("rns:fatigue", 0);

// 	s.propertyRegistry.registerEntityTypeDynamicProperties(data, EntityTypes.get("minecraft:player"));
// });

let playerHurt = [];
let playerInBed =  {};
let rollCooldown = [];
let rollTime = [];
let staminaDuration = [];
let stamina_is_cooldown = [];
let player_on_ground = [];
var fatigue_data = world.scoreboard.getObjective("rns:fatigue");
if(fatigue_data == undefined){
	fatigue_data = world.scoreboard.addObjective("rns:fatigue", "fatigue");
}

let manuver = {};

let ui_temp = {};
let current_player_speed = {};
let time_buffer = world.getAbsoluteTime();

var roll_path = [
	"",
	"textures/ui/roll/default/cooldown_",
	"textures/ui/roll/dungeon/cooldown_"
]

var stamina_path = [
	"",
	"textures/ui/stamina/default/bar_",
	"textures/ui/stamina/zelda/bar_"
]
var edge_climb_path = [
	"textures/ui/edge_climb/bar_0",
	"textures/ui/edge_climb/bar_1",
	"textures/ui/edge_climb/bar_2",
	"textures/ui/edge_climb/bar_3",
	"textures/ui/edge_climb/bar_4",
	"textures/ui/edge_climb/bar_5",
	"textures/ui/edge_climb/bar_6",
	"textures/ui/edge_climb/bar_7",
	"textures/ui/edge_climb/bar_8",
	"textures/ui/edge_climb/bar_9",
	"textures/ui/edge_climb/bar_10",
	"textures/ui/edge_climb/bar_11",
	"textures/ui/edge_climb/bar_12",
	"textures/ui/edge_climb/bar_13",
	"textures/ui/edge_climb/bar_14",
	"textures/ui/edge_climb/bar_15",
	"textures/ui/edge_climb/bar_16"
]

let edge_climb_duration = {};
var edge_climb_motion = [
	[ 0, 0.5 ],
	[ 0, 0.6 ],
	[ 0.1, 0.5 ],
	[ 0.2, 0.3 ],
	[ 0.6, 0.1 ],
	[ 0.3, 0.0 ],
	[ 0.0, -0.1 ],
	[ 0, -0.1 ],
	[ 0, 0 ]
]

function setUI(playerData, path, id){
	if(!ui_temp[playerData.id]){
		ui_temp[playerData.id] = { [id]: path };
	}else{
		if(ui_temp[playerData.id][id] == path) return;
	}
//	console.warn(path)
	playerData.runCommandAsync("scriptevent ui_load:" + id + " " + path + id);
	ui_temp[playerData.id][id] = path;
}

system.runInterval(() => {
	let creative_player = world.getPlayers({gameMode: "creative"});
	let stamina_strength = Number(config["rns:stamina_strength"]);
	let stamina_exhausted = Number(config["rns:stamina_exhausted"])*0.5+0.5;
	let stamina_cooldown = Number(config["rns:stamina_cooldown"])*0.5+0.5;
	
	let stamina_style = Number(config["rns:stamina_style"]);
	let roll_style = Number(config["rns:roll_style"]);
	let time = system.currentTick;
	for(let playerData of world.getPlayers()){
		if(refresh_ui == true){
			if(roll_style > 0){
				if(rollCooldown[playerData.id] == undefined){
					setUI(playerData, roll_path[roll_style] + 15, "roll_update");
				}else{
					let cooldown_delta = time - rollCooldown[playerData.id];
					cooldown_delta = parseInt(cooldown_delta/6);
					if(cooldown_delta <= 15){
						setUI(playerData, roll_path[roll_style] + cooldown_delta, "roll_update");
					}else{
						setUI(playerData, roll_path[roll_style] + 15, "roll_update");
					}
				}
			}else{
				setUI(playerData, "", "roll_update");
			}
			
			
			if(stamina_style > 0){
				if(staminaDuration[playerData.id] != undefined){
					let cooldown_delta = (staminaDuration[playerData.id])/stamina_strength;
					cooldown_delta = 100-parseInt(Math.min(cooldown_delta, 1)*100);
					setUI(playerData, stamina_path[stamina_style]  + cooldown_delta, "stamina_update");
				}else{
					setUI(playerData, stamina_path[stamina_style]  + 100, "stamina_update");
				}
			}else{
				setUI(playerData, "", "stamina_update");
			}
			setUI(playerData, "", "edge_climb_update");
			if(config["rns:fatigue"] == "false") setUI(playerData, "", "fatigue_update");
		}
		//fatigue MAX 2048

		let fatigue = undefined;
		try{
			fatigue = fatigue_data.getScore(playerData);
		}catch(err){
			console.warn(playerData.name + " has no data, creating new data")
		}
		
		if(fatigue == undefined){
			fatigue_data.addScore(playerData, 0);
			fatigue = 0;
		}

		let multiplier_stamina = 1;
		// console.warn(fatigue)
		// console.warn(playerData.getBlockFromViewDirection({includePassableBlocks: false, maxDistance: 1}))

		if(config["rns:edge_climb"] == "true" && !playerData.isInWater){
			if(manuver[playerData.id] == undefined){
				manuver[playerData.id] = 0;
				edge_climb_duration[playerData.id] = 17;
			}
			if(manuver[playerData.id] > 0 && playerData.isOnGround) manuver[playerData.id] = 0;

			if(manuver[playerData.id] != 2 && !playerData.isOnGround) manuver[playerData.id] = playerData.isJumping ? 1 : 0;
			// console.warn(manuver[playerData.id])
			
			if(manuver[playerData.id] > 0 && edge_climb_duration[playerData.id] == 17){
				let rot = playerData.getRotation();
				let rot_y = rot.y / 180 * Math.PI;
				let dir = {
					x: -Math.sin(rot_y) * 0.5,
					y: 0,
					z: Math.cos(rot_y) * 0.5
				}
				let loc = playerData.getHeadLocation();
				loc.y += 2;
				loc = Vector.add(loc, dir);
				let block = playerData.getBlockFromViewDirection({includePassableBlocks: false, maxDistance: 1});
				let above_block = playerData.dimension.getBlockFromRay(loc, {x: 0, y: -1, z: 0}, {includePassableBlocks: false, maxDistance: 4});

				if(block == undefined && above_block != undefined){
					if(Math.abs((above_block.block.location.y + above_block.faceLocation.y) - (loc.y - 2)) < 0.5){
						// console.warn("is edge jump")
						edge_climb_duration[playerData.id] = 0;
						let i = 0;
						let stop_cast = time + 20;
						if(config["rns:stamina"] == "true") staminaDuration[playerData.id] += 10*stamina_exhausted;
						playerData.playAnimation("animation.humanoid.edge_climb", {
							controller: "climb_controller"
						});
						playerData.dimension.spawnParticle("rns:move_smoke", playerData.location);

						let runId = system.runInterval(()=>{
							playerData.applyKnockback(dir.x, dir.z, edge_climb_motion[i][0], edge_climb_motion[i][1]);
							playerData.addEffect("slow_falling", 1, {amplifier: 64, showParticles: false})
							if((playerData.location.y + 1) > (above_block.block.location.y + above_block.faceLocation.y)) i++;
							if(i > 8 || system.currentTick > stop_cast){
								system.clearRun(runId);
								let cooldownId = system.runInterval(()=>{
									setUI(playerData, edge_climb_path[edge_climb_duration[playerData.id]], "edge_climb_update");
									edge_climb_duration[playerData.id] = edge_climb_duration[playerData.id] + 1;
									if(edge_climb_duration[playerData.id] == 17){
										system.clearRun(cooldownId);
										setUI(playerData, "", "edge_climb_update");
									}
								});
							}
						});
					}
				}
			}
		}

		let is_creative = false;
		for(let player of creative_player){
			if(player.id == playerData.id){
				is_creative = true;
				break;
			}
		}
		if(!playerInBed[playerData.id]){
			playerInBed[playerData.id] = playerData.isSleeping;
		}else{
			if(playerData.isSleeping == true){
				system.runTimeout(()=>{
					playerInBed[playerData.id] = false;
				}, 2);
			}
		}

		if(world.getAbsoluteTime() - time_buffer > 1 && playerInBed[playerData.id]){
			staminaDuration[playerData.id] = 0;
		}



		if(!is_creative && config["rns:fatigue"] == "true"){
			if(world.getAbsoluteTime() - time_buffer > 1 && playerInBed[playerData.id]){
				// playerData.setDynamicProperty("rns:fatigue", fatigue - (world.getAbsoluteTime() - time_buffer) / 200);
				fatigue -= (world.getAbsoluteTime() - time_buffer) / 4;
				fatigue = Math.max(fatigue, 0);
				fatigue_data.setScore(playerData, fatigue);
				playerData.removeEffect("mining_fatigue");
				playerData.removeEffect("weakness");
			}

			if(fatigue/128 < 17){
				// playerData.setDynamicProperty("rns:fatigue", fatigue + 0.0005);
				fatigue += 1 / 2;
			}
			if(fatigue < 0){
				// playerData.setDynamicProperty("rns:fatigue", 0.0);
				fatigue = 0;
			} 
			if((time_buffer % 400 == 0 || (fatigue * 10) % 10 > 9.99)){
				setUI(playerData, "textures/ui/fatigue/fatigue_" + Math.min(parseInt(fatigue/128), 16), "fatigue_update");
			}
			if(fatigue/128 > 14 && playerData.getEffect("weakness") == undefined) playerData.addEffect("weakness", 600, {amplifier: parseInt(fatigue/128 - 14)});
			if(playerData.getEffect("mining_fatigue") == undefined){
				if(fatigue/128 > 12) playerData.addEffect("mining_fatigue", 600, {amplifier: parseInt(fatigue/128 - 12)});
				
			}else{
				// playerData.setDynamicProperty("rns:fatigue", fatigue + 0.001);
				fatigue += 2 / 2;
			}
			
			multiplier_stamina = Math.max((fatigue/128 - 8) / 4 , 0) + 1;
		}
		
		let is_riding = false;
		for(let ride of playerData.dimension.getEntities({ location: playerData.location, excludeTypes: ["minecraft:player"], maxDistance: 2})){
			if(ride.hasComponent("minecraft:rideable")){
				is_riding = true;
				break;
			}
		}
		
		let vel = playerData.getVelocity();
		let player_speed = playerData.getComponent("minecraft:movement");
		let vel_calc = Math.sqrt(Math.pow(vel.x, 2) + Math.pow(vel.z, 2));
		let is_sprint = playerData.isSprinting;
		let stamina_amplification = 1;

		if(playerData.isInWater) player_on_ground[playerData.id] = false;
		if(!player_on_ground[playerData.id] && playerData.isOnGround){
			player_on_ground[playerData.id] = true;
		}

		if(config["rns:stamina"] == "true"){
			let get_hunger = playerData.getEffect("hunger");
			let get_speed = playerData.getEffect("speed");

			if(config["rns:speedup"] == "true" && current_player_speed[playerData.id] != undefined && is_sprint){
				if(staminaDuration[playerData.id] < stamina_strength / 2.0){
					player_speed.setCurrentValue(current_player_speed[playerData.id] * ((1.75 - Math.min(staminaDuration[playerData.id] / stamina_strength * 2.0, 1.0) * 0.5)));
				}
			}else{
				current_player_speed[playerData.id] = player_speed.currentValue;
			}
			
			if(get_hunger != undefined){
				stamina_amplification += get_hunger.amplifier / 2;
			}
			if(get_speed != undefined){
				stamina_amplification /= 1 + get_speed.amplifier / 4;
			}
			if(staminaDuration[playerData.id] == undefined) staminaDuration[playerData.id] = 0;
			if(!is_creative){
				let use_stamina = false;
				
				if(playerData.isSwimming){
					if(stamina_is_cooldown[playerData.id]){
						playerData.applyKnockback(0, 0, 0, 0);
					}else{
						staminaDuration[playerData.id] += 0.5*stamina_exhausted*stamina_amplification * multiplier_stamina;
						if(config["rns:fatigue"] == "true"){
							// playerData.setDynamicProperty("rns:fatigue", fatigue + 0.0005);
							fatigue += 1 / 2;
						}
						use_stamina = true;
					}
				}else{
					if(!stamina_is_cooldown[playerData.id] && player_on_ground[playerData.id] && playerData.isJumping && !playerData.getEffect("jump_boost") && !is_riding){
						player_on_ground[playerData.id] = false;
						staminaDuration[playerData.id] += 10*stamina_exhausted*stamina_amplification * multiplier_stamina;
						if(config["rns:fatigue"] == "true"){
							// playerData.setDynamicProperty("rns:fatigue", fatigue + 0.005);
							fatigue += 5 / 2;
						}
						use_stamina = true;
					}
					
					if(is_sprint && !is_riding && playerData.isOnGround && !stamina_is_cooldown[playerData.id]){
						if(staminaDuration[playerData.id] < stamina_strength && !playerData.isSwimming){
							staminaDuration[playerData.id] += 1.5*stamina_exhausted*stamina_amplification * multiplier_stamina;
							if(config["rns:fatigue"] == "true"){
								// playerData.setDynamicProperty("rns:fatigue", fatigue + 0.0005);
								fatigue += 1 / 2;
							}
						}
						use_stamina = true;
					}

					if(stamina_is_cooldown[playerData.id] && playerData.isJumping && player_on_ground[playerData.id]) playerData.applyKnockback(0, 0, 0, -0.25);
				}

				if(!use_stamina && staminaDuration[playerData.id] > 0) staminaDuration[playerData.id] -= ((playerData.isSneaking) ? 1 : 0.5)*stamina_cooldown / multiplier_stamina; 
				
				if(staminaDuration[playerData.id] >= stamina_strength){
					staminaDuration[playerData.id] = stamina_strength;
					let speed_value = player_speed.defaultValue;
					let speed_decreased = speed_value/1.5;
					let runLoop = system.runInterval(() => {
						try{
							if(staminaDuration[playerData.id] < stamina_strength * 0.75){
								player_speed.setCurrentValue(speed_value);
								stamina_is_cooldown[playerData.id] = false;
								system.clearRun(runLoop);
							}else{
								player_speed.setCurrentValue(speed_decreased);
								stamina_is_cooldown[playerData.id] = true;
							}
						}catch(err){
							system.clearRun(runLoop);
						}
					});
				}
				if(staminaDuration[playerData.id] >= 0){
					let cooldown_delta = (staminaDuration[playerData.id])/stamina_strength;
					cooldown_delta = 100-parseInt(Math.min(cooldown_delta, 1)*100);
					if(stamina_style != 0) setUI(playerData, stamina_path[stamina_style]  + cooldown_delta, "stamina_update");
				}
				
				if(time % 40 == 0){
					fatigue = Math.ceil(fatigue);
					fatigue_data.setScore(playerData, Math.min(fatigue, 2048));
				}
			}
		}
		if(config["rns:roll"] == "false") continue;
		
		if(rollCooldown[playerData.id] != -1 && rollCooldown[playerData.id] != undefined){
			let cooldown_delta = time - rollCooldown[playerData.id];
			cooldown_delta = parseInt(cooldown_delta / 6 / multiplier_stamina * (parseInt(config["rns:roll_cooldown"]) + 1));
			if(cooldown_delta <= 15){
				if(roll_style != 0) setUI(playerData, roll_path[roll_style] + cooldown_delta, "roll_update");
			}else{
				rollCooldown[playerData.id] = -1;
			}
			return;
		}
		
		if(playerData.isInWater || stamina_is_cooldown[playerData.id]) return;
		if(playerData.isSneaking){
			if(rollTime[playerData.id] == -1 || rollTime[playerData.id] == undefined){
				rollTime[playerData.id] = time;
			}
		}else{
			playerData.isSneaking = false;
			if(rollTime[playerData.id] != -1 && rollTime[playerData.id] != undefined){
				if(time - rollTime[playerData.id] < 5){
					if(config["rns:edge_climb"] == "true") manuver[playerData.id] = 2;
					let rot = playerData.getRotation();
					let rot_temp = Math.floor((Math.atan2(vel.z, vel.x) * 180 / Math.PI - 90));
					if(rot_temp < -180){
						rot_temp = rot_temp+360;
					}
					let delta_rot = Math.abs(rot_temp - Math.floor((rot.y)))
					delta_rot = Math.min(delta_rot, (360-delta_rot))
					let power = 1;
					let strength = (1-Math.min(Math.abs(vel.y*3), 1))*2.3+0.2;
					let roll_anim = (delta_rot > 120) ? "animation.humanoid.roll_down" : "animation.humanoid.roll_up";
					let roll_anim_def = (delta_rot > 120) ? "animation.humanoid.roll_steady2" : "animation.humanoid.roll_steady";
					if(Math.abs(vel.x) + Math.abs(vel.z) > 0.01){
						let vel_up = vel_calc*2;
						playerData.applyKnockback(vel.x, vel.z, strength * ((delta_rot > 120) ? 1.5 : 1)*0.1*Number(config["rns:roll_horizontal_strength"]), ((delta_rot > 120) ? 0.2 : 0.4)*Number(config["rns:roll_vertical_strength"])*0.1);
					}else{
						rot.y = (rot.y + 45)* Math.PI/180;
						let velocity = {
							"x": (Math.cos(rot.y)-Math.sin(rot.y))*power,
							"y": 0,
							"z": (Math.sin(rot.y)+Math.cos(rot.y))*power,
						};
						playerData.applyKnockback(velocity.x, velocity.z, strength/2*0.1*Number(config["rns:roll_horizontal_strength"]), 0.6*Number(config["rns:roll_vertical_strength"])*0.1);
						roll_anim = "animation.humanoid.roll_up";
					}
					rollCooldown[playerData.id] = time;
					playerData.playAnimation(roll_anim_def, {
						blendOutTime: 0.3,
						controller: "roll_steady_controller"
					});
					playerData.playAnimation(roll_anim, {
						controller: "roll_controller"
					});
					playerData.playAnimation("animation.humanoid.roll_head", {
						blendOutTime: 1,
						controller: "roll_head_controller"
					});
					playerData.dimension.spawnParticle("rns:move_smoke", playerData.location);

					if(!playerData.getEffect("jump_boost") && config["rns:stamina"] == "true"){
						staminaDuration[playerData.id] += 10*stamina_exhausted*stamina_amplification * multiplier_stamina;
						if(config["rns:fatigue"] == "true"){
							// playerData.setDynamicProperty("rns:fatigue", fatigue + 0.005);
							fatigue += 5 / 2;
							fatigue = Math.ceil(fatigue);
							fatigue_data.setScore(playerData, Math.min(fatigue, 2048));
						}
					}
				}
				rollTime[playerData.id] = -1; 
			}
		}
	}
	if(refresh_ui == true) refresh_ui = false;
	time_buffer = world.getAbsoluteTime();
});

world.afterEvents.entityHurt.subscribe( s => {
	if(config["rns:slow_damage"] == "false") return;
	let player = s.hurtEntity;
	if(s.damageSource.cause == "magic" || playerHurt[player.id]) return;
	let player_speed = player.getComponent("minecraft:movement");
	let speed_value = player_speed.defaultValue;
	let speed_decreased = speed_value/1.5;
	let runLoop = system.runInterval(() => {
		player_speed.setCurrentValue(speed_decreased);
		playerHurt[player.id] = true;
	});
	let run = system.runInterval(() => {
		system.clearRun(runLoop);
		player_speed.setCurrentValue(speed_value);
		playerHurt[player.id] = false;
		system.clearRun(run);
	}, 30);
}, {entityTypes: [ "minecraft:player" ]});

world.afterEvents.entityHitBlock.subscribe( s => {
	if(config["rns:fatigue"] != "true") return;

	// let fatigue = s.damagingEntity.getDynamicProperty("rns:fatigue");
	// s.damagingEntity.setDynamicProperty("rns:fatigue", fatigue + 0.005);
	fatigue_data.addScore(s.damagingEntity, 5 / 2);
}, {entityTypes: [ "minecraft:player" ]});

world.afterEvents.entityHitEntity.subscribe( s => {
	if(config["rns:fatigue"] != "true") return;

	// let fatigue = s.damagingEntity.getDynamicProperty("rns:fatigue");
	// s.damagingEntity.setDynamicProperty("rns:fatigue", fatigue + 0.001);
	fatigue_data.addScore(s.damagingEntity, 2 / 2);
}, {entityTypes: [ "minecraft:player" ]});